""" This a collection of tools for SentryReporter and SentryScrubber aimed to
simplify work with several data structures.
"""
import re
from typing import Any, Callable, Dict, List, Optional, TypeVar

from faker import Faker

# Remove the substring like "Sentry is attempting to send 1 pending error messages"
_re_remove_sentry = re.compile(r'Sentry is attempting.*')


def get_first_item(items, default=None):
    return items[0] if items else default


def get_last_item(items, default=None):
    return items[-1] if items else default


def delete_item(d, key):
    if not d:
        return d

    if key in d:
        del d[key]
    return d


def get_value(d, key, default=None):
    return d.get(key, default) if d else default


def extract_dict(d, regex_key_pattern):
    if not d or not regex_key_pattern:
        return dict()

    matched_keys = [key for key in d if re.match(regex_key_pattern, key)]
    return {key: d[key] for key in matched_keys}


def modify_value(d, key, function):
    if not d or not key or not function:
        return d

    if key in d:
        d[key] = function(d[key])

    return d


T = TypeVar('T')


def distinct_by(items: Optional[List[T]], getter: Callable[[T], Any]) -> Optional[List[T]]:
    """This function removes all duplicates from a list of dictionaries. A duplicate
    here is a dictionary that have the same value of the given key.

    If no key field is presented in the dictionary, then the exception will be raised.

    Args:
        items: list of dictionaries
        getter: function that returns a key for the comparison

    Returns:
        Array of distinct items
    """

    if not items:
        return items

    distinct = {}
    for item in items:
        key = getter(item)
        if key not in distinct:
            distinct[key] = item
    return list(distinct.values())


def obfuscate_string(s: str, part_of_speech: str = 'noun') -> str:
    """Obfuscate string by replacing it with random word.

    The same random words will be generated for the same given strings.
    """
    faker = Faker(locale='en_US')
    faker.seed_instance(s)
    return faker.word(part_of_speech=part_of_speech)


def order_by_utc_time(breadcrumbs: Optional[List[Dict]], key: str = 'timestamp'):
    """ Order breadcrumbs by timestamp in ascending order.

    Args:
        breadcrumbs: List of breadcrumbs
        key: Field name that will be used for sorting

    Returns:
        Ordered list of breadcrumbs
    """
    if not breadcrumbs:
        return breadcrumbs

    return list(sorted(breadcrumbs, key=lambda breadcrumb: breadcrumb[key]))



from unittest.mock import Mock, patch


@patch('importlib.metadata.version', new=Mock(return_value="2.0.0"))
def test_version_success():
    """ Test version function """
    from sentry_scrubber import get_version
    assert get_version() == "2.0.0"


@patch('importlib.metadata.version', new=Mock(side_effect=ImportError))
def test_version_import_error():
    """ Test version function with ImportError """
    from sentry_scrubber import get_version
    assert get_version() == "1.0.0"



import pytest

from scrubber import SentryScrubber


@pytest.fixture
def scrubber():
    return SentryScrubber()


FOLDERS_POSITIVE_MATCH = [
    '/home/username/some/',
    '/usr/local/path/',
    '/Users/username/some/',
    '/users/username/some/long_path',
    '/home/username/some/',
    '/data/media/username3/some/',
    'WINNT\\Profiles\\username\\some',
    'Documents and Settings\\username\\some',
    'C:\\Users\\Some User\\',
    'C:\\Users\\USERNAM~1\\',

    # double slashes (could be present as errors during a serialisation)
    'C:\\\\Users\\\\username\\\\',
    '//home//username//some//',

]

FOLDERS_NEGATIVE_MATCH = [
    '',
    'any text',
    '/home//some/',
]


@pytest.mark.parametrize('folder', FOLDERS_NEGATIVE_MATCH)
def test_patterns_folders_negative_match(folder: str, scrubber: SentryScrubber):
    """ Test that the scrubber does not match folders """
    assert not any(regex.search(folder) for regex in scrubber.re_folders)


@pytest.mark.parametrize('folder', FOLDERS_POSITIVE_MATCH)
def test_patterns_folders_positive_match(folder: str, scrubber: SentryScrubber):
    """ Test that the scrubber matches folders """
    assert any(regex.search(folder) for regex in scrubber.re_folders)


IP_POSITIVE_MATCH = [
    '127.0.0.1',
    '0.0.0.1',
    '0.100.0.1',
    '(0.100.0.1)'
]

IP_NEGATIVE_MATCH = [
    '0.0.0',
    '0.0.0.0.0',
    '0.1000.0.1',
    '0.a.0.1',
    '0123.0.0.1',
    '03.0.0.1234',
    'a0.0.0.1',
]


@pytest.mark.parametrize('ip', IP_NEGATIVE_MATCH)
def test_patterns_ip_negative_match(ip: str, scrubber: SentryScrubber):
    """ Test that the scrubber does not match IPs """
    assert not scrubber.re_ip.search(ip)


@pytest.mark.parametrize('ip', IP_POSITIVE_MATCH)
def test_patterns_ip_positive_match(ip: str, scrubber: SentryScrubber):
    """ Test that the scrubber matches IPs """
    assert scrubber.re_ip.search(ip)


HASH_POSITIVE_MATCH = [
    '3030303030303030303030303030303030303030',
    '0a30303030303030303030303030303030303030',
    'hash:3030303030303030303030303030303030303030'
]

HASH_NEGATIVE_MATCH = [
    '0a303030303030303030303030303030303030300',
    '0a3030303030303030303303030303030303030',
    'z030303030303030303030303030303030303030'
]


@pytest.mark.parametrize('h', HASH_NEGATIVE_MATCH)
def test_patterns_hash_negative_match(h: str, scrubber: SentryScrubber):
    """ Test that the scrubber does not match hashes """
    assert not scrubber.re_hash.search(h)


@pytest.mark.parametrize('h', HASH_POSITIVE_MATCH)
def test_patterns_hash_positive_match(h: str, scrubber: SentryScrubber):
    """ Test that the scrubber scrub hashes """
    assert scrubber.re_hash.search(h)


def test_scrub_path_negative_match(scrubber: SentryScrubber):
    """ Test that the scrubber does not scrub paths """
    assert scrubber.scrub_text('/usr/local/path/') == '/usr/local/path/'
    assert scrubber.scrub_text('some text') == 'some text'

    assert not scrubber.sensitive_occurrences


def test_dict_markers_to_scrub(scrubber: SentryScrubber):
    scrubber.dict_markers_to_scrub = {'marker': 'top secret'}

    event = {
        'not secret information': 'any',
        'suspicious': {
            'information': 'but not secret',
            'just contains': 'top secret',
        },
        'secret information': {
            'marker': 'top secret',
            'any': 'information',
        }
    }

    actual = scrubber.scrub_event(event)
    expected = {
        'not secret information': 'any',
        'suspicious': {
            'information': 'but not secret',
            'just contains': 'top secret'
        },
        'secret information': {}
    }

    assert actual == expected


def test_scrub_path_positive_match(scrubber: SentryScrubber):
    """ Test that the scrubber scrubs paths """
    assert scrubber.scrub_text('/users/user/apps') == '/users/<boot>/apps'
    assert 'user' in scrubber.sensitive_occurrences

    assert scrubber.scrub_text('/users/username/some/long_path') == '/users/<highlight>/some/long_path'
    assert 'username' in scrubber.sensitive_occurrences


def test_scrub_text_ip_negative_match(scrubber: SentryScrubber):
    """ Test that the scrubber does not scrub IPs """
    assert scrubber.scrub_text('127.0.0.1') == '127.0.0.1'
    assert scrubber.scrub_text('0.0.0') == '0.0.0'


def test_scrub_text_ip_positive_match(scrubber: SentryScrubber):
    """ Test that the scrubber scrubs IPs """
    assert scrubber.scrub_text('0.0.0.1') == '<IP>'
    assert scrubber.scrub_text('0.100.0.1') == '<IP>'

    assert not scrubber.sensitive_occurrences


def test_scrub_text_hash_negative_match(scrubber: SentryScrubber):
    """ Test that the scrubber does not scrub hashes """
    too_long_hash = '1' * 41
    assert scrubber.scrub_text(too_long_hash) == too_long_hash
    too_short_hash = '2' * 39
    assert scrubber.scrub_text(too_short_hash) == too_short_hash


def test_scrub_text_hash_positive_match(scrubber: SentryScrubber):
    """ Test that the scrubber scrubs hashes """
    assert scrubber.scrub_text('3' * 40) == '<hash>'
    assert scrubber.scrub_text('hash:' + '4' * 40) == 'hash:<hash>'

    assert not scrubber.sensitive_occurrences


def test_scrub_text_complex_string(scrubber):
    """ Test that the scrubber scrubs complex strings """
    source = (
        'this is a string that has been sent from '
        '192.168.1.1(3030303030303030303030303030303030303030) '
        'located at usr/someuser/path on '
        "someuser's machine(someuser_with_postfix)"
    )

    actual = scrubber.scrub_text(source)

    assert actual == ('this is a string that has been sent from '
                      '<IP>(<hash>) '
                      'located at usr/<effect>/path on '
                      "<effect>'s machine(someuser_with_postfix)")

    assert 'someuser' in scrubber.sensitive_occurrences
    assert scrubber.scrub_text('someuser') == '<effect>'


def test_scrub_simple_event(scrubber):
    """ Test that the scrubber scrubs simple events """
    assert scrubber.scrub_event(None) is None
    assert scrubber.scrub_event({}) == {}
    assert scrubber.scrub_event({'some': 'field'}) == {'some': 'field'}


def test_scrub_event(scrubber):
    """ Test that the scrubber scrubs events """
    event = {
        'the very first item': 'username',
        'server_name': 'userhost',
        'contexts': {
            'reporter': {
                'any': {
                    'USERNAME': 'User Name',
                    'USERDOMAIN_ROAMINGPROFILE': 'userhost',
                    'PATH': '/users/username/apps',
                    'TMP_WIN': r'C:\Users\USERNAM~1\AppData\Local\Temp',
                    'USERDOMAIN': ' USER-DOMAIN',  # it is a corner case when there is a space before a text
                    'COMPUTERNAME': 'Computer name',
                },
                'stacktrace': [
                    'Traceback (most recent call last):',
                    'File "/Users/username/Tribler/tribler/src/tribler-gui/tribler_gui/"',
                ],
                'sysinfo': {'sys.path': ['/Users/username/Tribler/', '/Users/username/', '.']},
            }
        },
        'extra': {'sys_argv': ['/Users/username/Tribler']},
        'logentry': {'message': 'Exception with username', 'params': ['Traceback File: /Users/username/Tribler/']},
        'breadcrumbs': {
            'values': [
                {'type': 'log', 'message': 'Traceback File: /Users/username/Tribler/', 'timestamp': '1'},
                {'type': 'log', 'message': 'IP: 192.168.1.1', 'timestamp': '2'},
            ]
        },
    }
    assert scrubber.scrub_event(event) == {
        'the very first item': '<highlight>',
        'server_name': '<kid>',
        'contexts': {
            'reporter': {
                'any': {
                    'USERNAME': '<conference>',
                    'USERDOMAIN_ROAMINGPROFILE': '<protection>',
                    'PATH': '/users/<highlight>/apps',
                    'TMP_WIN': 'C:\\Users\\<restaurant>\\AppData\\Local\\Temp',
                    'USERDOMAIN': '<tune>',
                    'COMPUTERNAME': '<lady>',
                },
                'stacktrace': [
                    'Traceback (most recent call last):',
                    'File "/Users/<highlight>/Tribler/tribler/src/tribler-gui/tribler_gui/"',
                ],
                'sysinfo': {
                    'sys.path': [
                        '/Users/<highlight>/Tribler/',
                        '/Users/<highlight>/',
                        '.',
                    ]
                },
            },
        },
        'logentry': {
            'message': 'Exception with <highlight>',
            'params': ['Traceback File: /Users/<highlight>/Tribler/'],
        },
        'extra': {'sys_argv': ['/Users/<highlight>/Tribler']},
        'breadcrumbs': {
            'values': [
                {
                    'type': 'log',
                    'message': 'Traceback File: /Users/<highlight>/Tribler/',
                    'timestamp': '1',
                },
                {'type': 'log', 'message': 'IP: <IP>', 'timestamp': '2'},
            ]
        },
    }


def test_entities_recursively(scrubber):
    """ Test that the scrubber scrubs entities recursively """

    # positive
    assert scrubber.scrub_entity_recursively(None) is None
    assert scrubber.scrub_entity_recursively({}) == {}
    assert scrubber.scrub_entity_recursively([]) == []
    assert scrubber.scrub_entity_recursively('') == ''
    assert scrubber.scrub_entity_recursively(42) == 42

    event = {
        'some': {
            'value': [
                {
                    'path': '/Users/username/Tribler'
                }
            ]
        }
    }
    assert scrubber.scrub_entity_recursively(event) == {
        'some': {'value': [{'path': '/Users/<highlight>/Tribler'}]}
    }
    # stop on depth
    assert scrubber.scrub_entity_recursively(event) != event
    assert scrubber.scrub_entity_recursively(event, depth=2) == event


def test_scrub_unnecessary_fields(scrubber):
    """ Test that the scrubber scrubs unnecessary fields """
    # default
    assert scrubber.scrub_event({'default': 'field'}) == {'default': 'field'}

    # custom
    custom_scrubber = SentryScrubber()
    custom_scrubber.event_fields_to_cut = {'new', 'default'}
    assert custom_scrubber.scrub_event({'default': 'event', 'new': 'field', 'modules': {}}) == {'modules': {}}


def test_scrub_text_none(scrubber):
    assert scrubber.scrub_text(None) is None


def test_scrub_dict(scrubber):
    assert scrubber.scrub_entity_recursively(None) is None
    assert scrubber.scrub_entity_recursively({}) == {}

    assert scrubber.scrub_entity_recursively({'key': [1]}) == {'key': [1]}  # non-string values should not lead to error

    given = {'PATH': '/home/username/some/', 'USERDOMAIN': 'UD', 'USERNAME': 'U', 'REPEATED': 'user username UD U',
             'key': ''}
    assert scrubber.scrub_entity_recursively(given) == {'PATH': '/home/<highlight>/some/',
                                                        'REPEATED': 'user <highlight> <school> <night>',
                                                        'USERDOMAIN': '<school>',
                                                        'USERNAME': '<night>',
                                                        'key': '<dress>'}

    assert 'username' in scrubber.sensitive_occurrences
    assert 'UD' in scrubber.sensitive_occurrences
    assert 'U' in scrubber.sensitive_occurrences
    assert '' not in scrubber.sensitive_occurrences


def test_scrub_list(scrubber):
    assert scrubber.scrub_entity_recursively(None) is None
    assert scrubber.scrub_entity_recursively([]) == []

    assert scrubber.scrub_entity_recursively(['/home/username/some/']) == ['/home/<highlight>/some/']
    assert 'username' in scrubber.sensitive_occurrences



import pytest

from utils import delete_item, distinct_by, extract_dict, get_first_item, get_last_item, get_value, modify_value, \
    obfuscate_string, order_by_utc_time


def test_first():
    assert get_first_item(None, '') == ''
    assert get_first_item([], '') == ''
    assert get_first_item(['some'], '') == 'some'
    assert get_first_item(['some', 'value'], '') == 'some'

    assert get_first_item((), '') == ''
    assert get_first_item(('some', 'value'), '') == 'some'

    assert get_first_item(None, None) is None


def test_last():
    assert get_last_item(None, '') == ''
    assert get_last_item([], '') == ''
    assert get_last_item(['some'], '') == 'some'
    assert get_last_item(['some', 'value'], '') == 'value'

    assert get_last_item((), '') == ''
    assert get_last_item(('some', 'value'), '') == 'value'

    assert get_last_item(None, None) is None


def test_delete():
    assert delete_item({}, None) == {}

    assert delete_item({'key': 'value'}, None) == {'key': 'value'}
    assert delete_item({'key': 'value'}, 'missed_key') == {'key': 'value'}
    assert delete_item({'key': 'value'}, 'key') == {}


def test_modify():
    assert modify_value(None, None, None) is None
    assert modify_value({}, None, None) == {}
    assert modify_value({}, '', None) == {}

    assert modify_value({}, 'key', lambda value: '') == {}
    assert modify_value({'a': 'b'}, 'key', lambda value: '') == {'a': 'b'}
    assert modify_value({'a': 'b', 'key': 'value'}, 'key', lambda value: '') == {'a': 'b', 'key': ''}


def test_safe_get():
    assert get_value(None, None, None) is None
    assert get_value(None, None, {}) == {}

    assert get_value(None, 'key', {}) == {}

    assert get_value({'key': 'value'}, 'key', {}) == 'value'
    assert get_value({'key': 'value'}, 'key1', {}) == {}


def test_distinct_none():
    # Test distinct_by with None
    assert distinct_by(None, lambda b: (b["timestamp"], b["message"])) is None


def test_distinct():
    # Test distinct_by with default getter
    values = [
        {'message': 'message 1', 'timestamp': 'timestamp 1', 'id': '1'},
        {'message': 'message 1', 'timestamp': 'timestamp 1', 'id': '2'},
        {'message': 'message 2', 'timestamp': 'timestamp 2', 'id': '3'}
    ]

    expected = [
        {'message': 'message 1', 'timestamp': 'timestamp 1', 'id': '1'},
        {'message': 'message 2', 'timestamp': 'timestamp 2', 'id': '3'}
    ]
    assert distinct_by(values, lambda b: (b["timestamp"], b["message"])) == expected


def test_distinct_key_error():
    # Test distinct_by with missing key in getter
    values = [
        {'message': 'message 1', },
    ]
    with pytest.raises(KeyError):
        distinct_by(values, lambda b: (b["timestamp"], b["message"]))


def test_distinct_none_in_list():
    # Test distinct_by with None in list
    values = [None]
    with pytest.raises(TypeError):
        distinct_by(values, lambda b: (b["timestamp"], b["message"]))


FORMATTED_VERSIONS = [
    (None, None),
    ('', ''),
    ('7.6.0', '7.6.0'),
    ('7.6.0-GIT', 'dev'),  # version from developers machines
    ('7.7.1-17-gcb73f7baa', '7.7.1'),  # version from deployment tester
    ('7.7.1-RC1-10-abcd', '7.7.1-RC1'),  # release candidate
    ('7.7.1-exp1-1-abcd ', '7.7.1-exp1'),  # experimental versions
    ('7.7.1-someresearchtopic-7-abcd ', '7.7.1-someresearchtopic'),
]


def test_extract_dict():
    assert not extract_dict(None, None)

    assert extract_dict({}, '') == {}
    assert extract_dict({'k': 'v', 'k1': 'v1'}, r'\w$') == {'k': 'v'}


OBFUSCATED_STRINGS = [
    ('', 'dress'),
    ('any', 'challenge'),
    ('string', 'quality'),
]


@pytest.mark.parametrize('given, expected', OBFUSCATED_STRINGS)
def test_obfuscate_string(given, expected):
    assert obfuscate_string(given) == expected


def test_order_by_utc_time():
    # Test order by timestamp
    breadcrumbs = [
        {
            "timestamp": "2016-04-20T20:55:53.887Z",
            "message": "3",
        },
        {
            "timestamp": "2016-04-20T20:55:53.845Z",
            "message": "1",
        },
        {
            "timestamp": "2016-04-20T20:55:53.847Z",
            "message": "2",
        },
    ]
    ordered_breadcrumbs = order_by_utc_time(breadcrumbs)
    messages = [d['message'] for d in ordered_breadcrumbs]
    assert messages == ['1', '2', '3']


def test_order_by_utc_time_empty_breadcrumbs():
    # Test empty breadcrumbs
    assert not order_by_utc_time(None)



import re
from typing import Any, Dict, List, Optional, Union

from sentry_scrubber.utils import delete_item, obfuscate_string

DEFAULT_EXCLUSIONS = {'local', '127.0.0.1'}

DEFAULT_KEYS_FOR_SCRUB = {'USERNAME', 'USERDOMAIN', 'server_name', 'COMPUTERNAME', 'key'}

# https://en.wikipedia.org/wiki/Home_directory
DEFAULT_HOME_FOLDERS = {
    'users',
    'usr',
    'home',
    'u01',
    'var',
    r'data\/media',
    r'WINNT\\Profiles',
    'Documents and Settings',
    'Users',
}


class SentryScrubber:
    """This class is responsible for scrubbing all sensitive
    and unnecessary information from Sentry events.
    """

    def __init__(
            self,
            home_folders: Optional[set] = None,
            dict_keys_for_scrub: Optional[set] = None,
            dict_markers_to_scrub: Optional[dict] = None,
            exclusions: Optional[set] = None,
            scrub_ip: bool = True,
            scrub_hash: bool = True,
    ):
        """
        Initializes the SentryScrubber with configurable parameters.

        Args:
            home_folders (Optional[set]): Set of home directory names to target for scrubbing.
            dict_keys_for_scrub (Optional[set]): Set of dictionary keys whose values should be scrubbed.
            dict_markers_to_scrub (Optional[dict]): Dictionary markers that indicate values to scrub.
            exclusions (Optional[set]): Set of values to exclude from scrubbing.
            scrub_ip (bool): Flag to enable or disable IP scrubbing. Defaults to True.
            scrub_hash (bool): Flag to enable or disable hash scrubbing. Defaults to True.

        Example:
            >>> scrubber = SentryScrubber(scrub_ip=False)
            >>> scrubbed_event = scrubber.scrub_event(event)
        """
        self.home_folders = home_folders or DEFAULT_HOME_FOLDERS
        self.dict_keys_for_scrub = dict_keys_for_scrub or DEFAULT_KEYS_FOR_SCRUB
        self.dict_markers_to_scrub = dict_markers_to_scrub or {}
        self.event_fields_to_cut = set()
        self.exclusions = exclusions or DEFAULT_EXCLUSIONS
        self.scrub_ip = scrub_ip
        self.scrub_hash = scrub_hash

        # this is the dict (key: sensitive_info, value: placeholder)
        self.sensitive_occurrences = {}

        # placeholders
        self.create_placeholder = lambda text: f'<{text}>'
        self.hash_placeholder = self.create_placeholder('hash')
        self.ip_placeholder = self.create_placeholder('IP')

        # compiled regular expressions
        self.re_folders = set()
        self.re_ip = None
        self.re_hash = None

        self._compile_re()

    def _compile_re(self):
        """
        Compiles all necessary regular expressions for scrubbing.

        Compiled Patterns:
            - Folder paths based on `home_folders`.
            - IP addresses if `scrub_ip` is enabled.
            - Hashes if `scrub_hash` is enabled.

        Example:
            >>> scrubber = SentryScrubber()
            >>> scrubber._compile_re()
        """
        slash = r'[/\\]'
        for folder in self.home_folders:
            for separator in [slash, slash * 2]:
                folder_pattern = rf'(?<={folder}{separator})[\w\s~]+(?={separator})'
                self.re_folders.add(re.compile(folder_pattern, re.I))

        if self.scrub_ip:
            self.re_ip = re.compile(r'(?<!\.)\b(\d{1,3}\.){3}\d{1,3}\b(?!\.)', re.I)
        if self.scrub_hash:
            self.re_hash = re.compile(r'\b[0-9a-f]{40}\b', re.I)

    def scrub_event(self, event, _=None):
        """
        Main method to scrub a Sentry event by removing sensitive and unnecessary information.

        Args:
            event (dict): A Sentry event represented as a dictionary.
            _ (Any, optional): Unused parameter for compatibility. Defaults to None.

        Returns:
            dict: The scrubbed Sentry event.

        Example:
            >>> scrubber = SentryScrubber()
            >>> scrubbed = scrubber.scrub_event(event)
        """
        if not event:
            return event

        # remove unnecessary fields
        for field_name in self.event_fields_to_cut:
            delete_item(event, field_name)

        # remove sensitive information
        scrubbed_event = self.scrub_entity_recursively(event)

        # this second call is necessary to complete the entities scrubbing
        # which were found at the end of the previous call
        scrubbed_event = self.scrub_entity_recursively(scrubbed_event)

        return scrubbed_event

    def scrub_text(self, text):
        """
        Replaces all sensitive information in the given text with corresponding placeholders.

        Sensitive Information:
            - IP addresses
            - User Names
            - 40-character hashes

        Args:
            text (str): The text to scrub.

        Returns:
            str: The scrubbed text.

        Example:
            >>> scrubber = SentryScrubber()
            >>> scrubbed_text = scrubber.scrub_text("User john_doe with IP 192.168.1.1 logged in.")
            >>> print(scrubbed_text)
            "User <hash> with IP <IP> logged in."
        """
        if text is None:
            return text

        def scrub_username(m):
            user_name = m.group(0)
            if user_name in self.exclusions:
                return user_name
            fake_username = obfuscate_string(user_name)
            placeholder = self.create_placeholder(fake_username)
            self.add_sensitive_pair(user_name, placeholder)
            return placeholder

        for regex in self.re_folders:
            text = regex.sub(scrub_username, text)

        if self.scrub_ip and self.re_ip:
            # cut an IP
            def scrub_ip(m):
                return self.ip_placeholder if m.group(0) not in self.exclusions else m.group(0)

            text = self.re_ip.sub(scrub_ip, text)

        if self.scrub_hash and self.re_hash:
            # cut hash
            text = self.re_hash.sub(self.hash_placeholder, text)

        # replace all sensitive occurrences in the whole string
        if self.sensitive_occurrences:
            escaped_sensitive_occurrences = [re.escape(user_name) for user_name in self.sensitive_occurrences]
            pattern = r'([^<]|^)\b(' + '|'.join(escaped_sensitive_occurrences) + r')\b'

            def scrub_value(m):
                if m.group(2) not in self.sensitive_occurrences:
                    return m.group(0)
                return m.group(1) + self.sensitive_occurrences[m.group(2)]

            text = re.sub(pattern, scrub_value, text)

        return text

    def scrub_entity_recursively(self, entity: Union[str, Dict, List, Any], depth=10):
        """
        Recursively traverses an entity to remove all sensitive information.

        Supports:
            1. Strings
            2. Dictionaries
            3. Lists

        All other data types are skipped.

        Args:
            entity (Union[str, Dict, List, Any]): The entity to scrub.
            depth (int, optional): The recursion depth limit. Defaults to 10.

        Returns:
            Union[str, Dict, List, Any]: The scrubbed entity.

        Example:
            >>> scrubber = SentryScrubber()
            >>> scrubbed = scrubber.scrub_entity_recursively(event_dict)
        """
        if depth < 0 or not entity:
            return entity

        depth -= 1

        if isinstance(entity, str):
            return self.scrub_text(entity)

        if isinstance(entity, list):
            return [self.scrub_entity_recursively(item, depth) for item in entity]

        if isinstance(entity, dict):
            result = {}
            for key, value in entity.items():
                if marker_value := self.dict_markers_to_scrub.get(key):
                    if value == marker_value:
                        result = {}
                        break

                if key in self.dict_keys_for_scrub and isinstance(value, str):
                    value = value.strip()
                    fake_value = obfuscate_string(value)
                    placeholder = self.create_placeholder(fake_value)
                    self.add_sensitive_pair(value, placeholder)
                    result[key] = placeholder
                else:
                    result[key] = self.scrub_entity_recursively(value, depth)
            return result

        return entity

    def add_sensitive_pair(self, text, placeholder):
        """
        Adds a sensitive text and its corresponding placeholder to the occurrences dictionary.

        Args:
            text (str): The sensitive text to be replaced.
            placeholder (str): The placeholder to replace the sensitive text.

        Example:
            >>> scrubber = SentryScrubber()
            >>> scrubber.add_sensitive_pair("john_doe", "<hashed_username>")
        """
        if not (text and text.strip()):  # Avoid replacing empty substrings
            return

        if text in self.sensitive_occurrences:
            return

        self.sensitive_occurrences[text] = placeholder
